;; Generated package description from f90-interface-browser.el  -*- no-byte-compile: t -*-
(define-package "f90-interface-browser" "1.1.0.20221221.75553" "Parse and browse f90 interfaces" '((cl-lib "0.7")) :commit "4e192f6e03bc96f996923b307248b7d5b1d398fa" :authors '(("Lawrence Mitchell" . "wence@gmx.li")) :maintainer '("Lawrence Mitchell" . "wence@gmx.li") :url "https://github.com/wence-/f90-iface/")
